/*-
 * Copyright (c) 2014 European Spallation Source ERIC.
 * Copyright (c) 2014 Cosylab d.d.
 *
 * This file is part of Naming Service.
 * Naming Service is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free
 * Software Foundation, either version 2 of the License, or any newer version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program. If not, see https://www.gnu.org/licenses/gpl-2.0.txt
 */

package org.openepics.names.old.model;

import com.google.common.base.Preconditions;

import javax.annotation.Nullable;
import javax.persistence.Column;
import javax.persistence.Entity;
import javax.persistence.JoinColumn;
import javax.persistence.ManyToOne;
import javax.persistence.Table;

import org.openepics.names.repository.model.Persistable;

import java.util.Date;

/**
 * A revision of a Device entity representing its state at some point in time.
 *
 * @author Marko Kolar
 */
@Entity
@Table(name = "devicerevision")
public class DeviceRevision extends Persistable {

    @JoinColumn(name = "device_id")
    private @ManyToOne Device device;

    @JoinColumn(name = "requestedby_id")
    private @ManyToOne UserAccount requestedBy;

    @Column(name = "requestdate")
    private Date requestDate;

    private boolean deleted;

    @JoinColumn(name = "section_id")
    private @ManyToOne NamePart section;

    @JoinColumn(name = "devicetype_id")
    private @ManyToOne NamePart deviceType;

    @Column(name = "instanceindex")
    private @Nullable String instanceIndex;

    @Column(name = "conventionname")
    private String conventionName;

    @Column(name = "conventionnameeqclass")
    private String conventionNameEqClass;

    @Column(name = "additionalinfo")
    private @Nullable String additionalInfo;

    @Column(name = "processorcomment")
    private @Nullable String processorComment;

    protected DeviceRevision() {}

    /**
     * Constructs a new device revision entity (data object with relations).
     *
     * @param device the device the revision pertains to
     * @param requestDate the time when the revision was proposed
     * @param requestedBy the user that proposed the revision. Null if the revision was generated by an automated
     * process.
     * @param deleted a flag signifying that the revision represents deletion of the device
     * @param section the section containing the device
     * @param deviceType the type of the device
     * @param instanceIndex an additional identifier that, in combination with other attributes, determine the unique
     * convention name of the device. Null if omitted.
     * @param conventionName the full name of the device in accordance with the naming convention
     * @param conventionNameEqClass the representative of the equivalence class the convention name belongs to. This is
     * used to ensure uniqueness of convention names when treating similar looking names (for example, containing 0 vs.
     * O, 1 vs. l) as equal.
     * @param additionalInfo Additional information (description, comment etc) about the device.
     */
    public DeviceRevision(
            Device device, Date requestDate, @Nullable UserAccount requestedBy, boolean deleted,
            NamePart section, NamePart deviceType, @Nullable String instanceIndex, String conventionName,
            String conventionNameEqClass, @Nullable String additionalInfo) {

        this(device, requestDate, requestedBy, deleted, section, deviceType, instanceIndex, conventionName,
                conventionNameEqClass, additionalInfo, null);
    }

    /**
     * Constructs a new device revision entity (data object with relations).
     *
     * @param device the device the revision pertains to
     * @param requestDate the time when the revision was proposed
     * @param requestedBy the user that proposed the revision. Null if the revision was generated by an automated
     * process.
     * @param deleted a flag signifying that the revision represents deletion of the device
     * @param section the section containing the device
     * @param deviceType the type of the device
     * @param instanceIndex an additional identifier that, in combination with other attributes, determine the unique
     * convention name of the device. Null if omitted.
     * @param conventionName the full name of the device in accordance with the naming convention
     * @param conventionNameEqClass the representative of the equivalence class the convention name belongs to. This is
     * used to ensure uniqueness of convention names when treating similar looking names (for example, containing 0 vs.
     * O, 1 vs. l) as equal.
     * @param additionalInfo additional information (description, comment etc) about the device
     * @param commitMessage commit message
     */
    public DeviceRevision(
            Device device, Date requestDate, @Nullable UserAccount requestedBy, boolean deleted,
            NamePart section, NamePart deviceType, @Nullable String instanceIndex, String conventionName,
            String conventionNameEqClass, @Nullable String additionalInfo, @Nullable String commitMessage) {

        Preconditions.checkNotNull(device);
        Preconditions.checkNotNull(requestDate);
        Preconditions.checkNotNull(section);
        // not check deviceType as device may have section but no deviceType
        Preconditions.checkArgument(instanceIndex == null || !instanceIndex.isEmpty());
        Preconditions.checkArgument(conventionName != null && !conventionName.isEmpty());
        Preconditions.checkArgument(conventionNameEqClass != null);
        Preconditions.checkArgument(additionalInfo == null || !additionalInfo.isEmpty());
        this.device = device;
        this.requestedBy = requestedBy;
        this.requestDate = new Date(requestDate.getTime());
        this.deleted = deleted;
        this.section = section;
        this.deviceType = deviceType;
        this.instanceIndex = instanceIndex;
        this.conventionName = conventionName;
        this.conventionNameEqClass = conventionNameEqClass;
        this.additionalInfo=additionalInfo;
        this.processorComment=commitMessage;
    }

    /**
     * @return The device the revision pertains to.
     */
    public Device getDevice() { return device; }

    /**
     * @return The time when the revision was proposed.
     */
    public Date getRequestDate() {
        return requestDate != null ? new Date(requestDate.getTime()) : null;
    }

    /**
     * @return The user that proposed the revision. Null if the revision was generated by an automated process.
     */
    public @Nullable UserAccount getRequestedBy() { return requestedBy; }

    /**
     * @return A flag signifying that the revision represents deletion of the device.
     */
    public boolean isDeleted() { return deleted; }

    /**
     * @return The section containing the device.
     */
    public NamePart getSection() { return section; }

    /**
     * @return The type of the device.
     */
    public NamePart getDeviceType() { return deviceType; }

    /**
     * @return  An additional identifier that, in combination with other attributes,
     *          determine the unique convention name of the device. Null if omitted.
     */
    public @Nullable String getInstanceIndex() { return instanceIndex; }

    /**
     * @return The full name of the device in accordance with the naming convention.
     */
    public String getConventionName() { return conventionName; }

    /**
     * @return  The representative of the equivalence class the convention name belongs to.
     *          This is used to ensure uniqueness of convention names when treating similar
     *          looking names (for example, containing 0 vs. O, 1 vs. l) as equal.
     */
    public String getConventionNameEqClass() { return conventionNameEqClass; }

    /**
     * Sets the convention equivalent class
     * @param conventionNameEqClass the convention equivalent class to be set.
     */
    public void setConventionNameEqClass(String conventionNameEqClass){
       this.conventionNameEqClass= conventionNameEqClass;
    }

    /**
     * @return Additional information (description, comment etc) about the device.
     */
    public @Nullable String getAdditionalInfo(){ return additionalInfo;}

    @Nullable
    public String getProcessorComment() {
        return processorComment;
    }

    public void setProcessorComment(@Nullable String processorComment) {
        this.processorComment = processorComment;
    }

    public boolean supersede(DeviceRevision other) {
        return other == null
                || getId() > other.getId();
    }

}
