/*
 * Copyright (c) 2014 European Spallation Source ERIC.
 * Copyright (c) 2014 Cosylab d.d.
 *
 * This software is Copyright by the Board of Trustees of Michigan
 * State University (c) Copyright 2012.
 *
 * You may use this software under the terms of the GNU public license
 *  (GPL). The terms of this license are described at:
 *       http://www.gnu.org/licenses/gpl.txt
 *
 * Contact Information:
 *   Facilitty for Rare Isotope Beam
 *   Michigan State University
 *   East Lansing, MI 48824-1321
 *   http://frib.msu.edu
 */

package org.openepics.names.old.model;

import com.google.common.base.Preconditions;

import javax.annotation.Nullable;
import javax.persistence.Column;
import javax.persistence.Entity;
import javax.persistence.EnumType;
import javax.persistence.Enumerated;
import javax.persistence.JoinColumn;
import javax.persistence.ManyToOne;
import javax.persistence.Table;

import org.openepics.names.repository.model.Persistable;

import java.util.Date;

/**
 * A revision of a NamePart entity representing its state at some point in time.
 *
 * @author Vasu V
 * @author Marko Kolar
 * @author Karin Rathsman
 */
@Entity
@Table(name = "namepartrevision")
public class NamePartRevision extends Persistable {

    @JoinColumn(name = "namepart_id")
    private @ManyToOne NamePart namePart;

    @Column(name = "requestdate")
    private Date requestDate;

    @JoinColumn(name = "requestedby_id")
    private @ManyToOne @Nullable UserAccount requestedBy;

    @Column(name = "requestercomment")
    private @Nullable String requesterComment;

    private boolean deleted;

    @JoinColumn(name = "parent_id")
    private @ManyToOne @Nullable NamePart parent;

    private String name;

    private @Nullable String mnemonic;

    private @Nullable String description=null;

    @Column(name = "mnemoniceqclass")
    private @Nullable String mnemonicEqClass;

    @Enumerated(EnumType.STRING)
    private NamePartRevisionStatus status;

    @JoinColumn(name = "processedby_id")
    private @ManyToOne @Nullable UserAccount processedBy = null;

    @Column(name = "processdate")
    private @Nullable Date processDate = null;

    @Column(name = "processorcomment")
    private @Nullable String processorComment = null;

    protected NamePartRevision() {}

    /**
     * Constructs a new name part revision entity (data object with relations).
     *
     * @param namePart          the name part the revision pertains to
     * @param requestDate       the time when the revision was proposed
     * @param requestedBy       the user that proposed the revision.
     *                          Null if the revision was generated by an automated process.
     * @param requesterComment  the comment the user gave when proposing the revision. Null if no comment was given.
     * @param deleted           a flag signifying that the revision represents deletion of the name part
     * @param parent            the parent of this name part in the hierarchy. Null if at the top of the hierarchy.
     * @param name              is the full name of the part. Does not need to follow a convention.
     * @param mnemonic          is the short, mnemonic name of the part in accordance with the naming convention
     * @param description       is the (lengthy) description or comment of the part.
     * @param mnemonicEqClass   the representative of the equivalence class the mnemonic belongs to.
     *                          This is used to ensure uniqueness of mnemonics on certain levels when treating
     *                          similar looking names (for example, containing 0 vs. O, 1 vs. l) as equal.
     */
    public NamePartRevision(
            NamePart namePart, Date requestDate, @Nullable UserAccount requestedBy, @Nullable String requesterComment,
            boolean deleted, @Nullable NamePart parent, String name, String mnemonic,
            @Nullable String description, String mnemonicEqClass) {

        Preconditions.checkNotNull(namePart);
        Preconditions.checkNotNull(requestDate);
        Preconditions.checkArgument(requesterComment == null || !requesterComment.isEmpty());
        Preconditions.checkArgument(name != null && !name.isEmpty());
        this.namePart = namePart;
        this.requestDate = new Date(requestDate.getTime());
        this.requestedBy = requestedBy;
        this.requesterComment = requesterComment;
        this.deleted = deleted;
        this.parent = parent;
        this.name = name;
        this.mnemonic = mnemonic;
        this.description = description;
        this.mnemonicEqClass = mnemonicEqClass;
        this.status = NamePartRevisionStatus.PENDING;
    }

    /**
     * @return The name part the revision pertains to.
     */
    public NamePart getNamePart() { return namePart; }

    /**
     * Sets the status
     * @param status the status to be set
     */
    public void setStatus(NamePartRevisionStatus status){
        this.status = status;
    }

    /**
     *  @return The time when the revision was proposed.
     */
    public Date getRequestDate() {
        return requestDate != null ? new Date(requestDate.getTime()) : null;
    }

    /**
     *  @return The user that proposed the revision. Null if the revision was generated by an automated process.
     */
    public @Nullable UserAccount getRequestedBy() { return requestedBy; }

    /**
     *  @return The comment the user gave when proposing the revision. Null if no comment was given.
     */
    public @Nullable String getRequesterComment() { return requesterComment; }

    /**
     *  @return A flag signifying that the revision represents deletion of the name part.
     */
    public boolean isDeleted() { return deleted; }

    /**
     *  @return The parent of this name part in the hierarchy. Null if at the top of the hierarchy.
     */
    public @Nullable NamePart getParent() { return parent; }

    /**
     *  @return Full name of the part. Does not need to follow a convention.
     */
    public String getName() { return name; }

    /**
     *  @return The short, mnemonic name of the part in accordance with the naming convention.
     */
    public @Nullable String getMnemonic() { return mnemonic; }

    /**
     *  @return The description or other relevant information of the namepart. Optional.
     */
    public @Nullable String getDescription(){ return description; }

    /**
     *  @return     The representative of the equivalence class the mnemonic belongs to.
     *              This is used to ensure uniqueness of mnemonics on certain level when treating
     *              similar looking names (for example, containing 0 vs. O, 1 vs. l) as equal.
     */
    public @Nullable String getMnemonicEqClass() { return mnemonicEqClass; }

    public void setMnemonicEqClass(String mnemonicEqClass) { this.mnemonicEqClass = mnemonicEqClass; }

    /**
     *  @return The status of the name part in the request / approve workflow.
     */
    public NamePartRevisionStatus getStatus() { return status; }

    /**
     *  @return The time the revision was processed in the request / approve workflow. Null if still pending.
     */
    public @Nullable Date getProcessDate() {
        return processDate != null ? new Date(processDate.getTime()) : null;
    }

    /**
     * Sets the process date
     * @param date the date to be set
     */
    public void setProcessDate(Date date) {
        this.processDate = date != null ? new Date(date.getTime()) : null;
    }

    /**
     *  @return     The user who processed the revision in the request / approve workflow.
     *              Null if still pending or if the revision  @return was processed by an automated process.
     */
    public @Nullable UserAccount getProcessedBy() { return processedBy; }

    /**
     * Sets the process user
     * @param user the process user
     */
    public void setProcessedBy(UserAccount user){
        this.processedBy = user;
    }

    /**
     *  @return     The comment the user gave when processing the revision in the request / approve workflow.
     *              Null if still pending  @return or if no comment was given.
     */
    public @Nullable String getProcessorComment() { return processorComment; }

    /**
     * Sets the processorComment
     * @param message the message
     */
    public void setProcessorComment(String message){
        this.processorComment = message;
    }

    /**
     * Updates the revision's status in the request / approve workflow as either approved, rejected or canceled.
     * This completes the workflow and no further status changes are possible after that.
     *
     * @param status    the new status of the name part
     * @param date      the time the revision was processed
     * @param by        the user who processed the revision. Null if the revision was processed by an automated process.
     * @param comment   the comment the administrator user gave when processing the revision.
     *                  Null if no comment was given.
     *
     * @deprecated since version 3.3.2
     */
    @Deprecated
    public void updateAsProcessed(
            NamePartRevisionStatus status, Date date, @Nullable UserAccount by, @Nullable String comment) {

        Preconditions.checkState(this.status == NamePartRevisionStatus.PENDING);
        Preconditions.checkArgument(status != NamePartRevisionStatus.PENDING);
        Preconditions.checkNotNull(date);
        Preconditions.checkArgument(comment == null || !comment.isEmpty());
        this.status = status;
        this.processDate = new Date(date.getTime());
        this.processedBy = by;
        this.processorComment = comment;
    }

    /**
     * rollback processed status to pending.
     *
     * @deprecated since version 3.3.2
     */
    @Deprecated
    public void rollBackProcessed() {
        Preconditions.checkState(!this.status.equals(NamePartRevisionStatus.PENDING),"Status cannot be Pending");
        this.status=NamePartRevisionStatus.PENDING;
        this.processDate = null;
        this.processedBy = null;
        this.processorComment = null;
    }

}
